@@ -64,7 +64,7 @@ def match_tunes_and_chromaticities(
     dq1_target: float = None,
     dq2_target: float = None,
     varied_knobs: Sequence[str] = None,
-    telescopic_squeeze: bool = False,
+    telescopic_squeeze: bool = True,
     step: float = 1e-7,
     calls: int = 100,
     tolerance: float = 1e-21,
@@ -89,7 +89,7 @@ def match_tunes_and_chromaticities(
             could be ["kqf", "ksd", "kqf", "kqd"] as they are common names used for quadrupole and
             sextupole strengths (foc / defoc) in most examples.
         telescopic_squeeze (bool): LHC specific. If set to True, uses the (HL)LHC knobs for Telescopic
-            Squeeze configuration. Defaults to False, but will default to True in a later release.
+            Squeeze configuration. Defaults to True as of Run III.
         step (float): step size to use when varying knobs.
         calls (int): max number of varying calls to perform.
         tolerance (float): tolerance for successfull matching.
@@ -112,17 +112,17 @@ def match_tunes_and_chromaticities(
         madx.command.lmdif(calls=calls, tolerance=tolerance)
         madx.command.endmatch()
         logger.trace("Performing routine TWISS")
-        madx.twiss()  # prevents errors if the user forget to do so before querying tables
+        madx.twiss()  # prevents errors if the user forgets to TWISS before querying tables
 
-    if q1_target and q2_target and dq1_target and dq2_target:
+    if q1_target is not None and q2_target is not None and dq1_target is not None and dq2_target is not None:
         logger.info(
             f"Doing combined matching to Qx={q1_target}, Qy={q2_target}, "
-            f"dq={dq1_target}, dqy={dq2_target} for sequence '{sequence}'"
+            f"dqx={dq1_target}, dqy={dq2_target} for sequence '{sequence}'"
         )
         logger.trace(f"Vary knobs sent are {varied_knobs}")
         match(*varied_knobs, q1=q1_target, q2=q2_target, dq1=dq1_target, dq2=dq2_target)
 
-    elif q1_target and q2_target:
+    elif q1_target is not None and q2_target is not None:
         logger.info(f"Matching tunes to Qx={q1_target}, Qy={q2_target} for sequence '{sequence}'")
         logger.trace(f"Vary knobs sent are {varied_knobs[:2]}")
         match(*varied_knobs[:2], q1=q1_target, q2=q2_target)  # first two in varied_knobs are tune knobs
@@ -199,8 +199,6 @@ def get_closest_tune_approach(
         sequence,
         qx_target,
         qy_target,
-        # dq1,  # remove?
-        # dq2,  # remove?
         varied_knobs=varied_knobs,
         step=step,
         calls=calls,
@@ -211,8 +209,8 @@ def get_closest_tune_approach(
     dqmin = madx.table.summ.q1[0] - madx.table.summ.q2[0] - (int(q1) - int(q2))
 
     logger.info("Restoring saved knobs")
-    for knob, knob_value in saved_knobs.items():
-        madx.globals[knob] = knob_value
+    with madx.batch():
+        madx.globals.update(saved_knobs)
     madx.twiss()
 
     return abs(dqmin)
