@@ -6,23 +6,19 @@ Created on 2020.02.03
 :author: Felix Soubelet (felix.soubelet@cern.ch)
 
 A module with functions to perform MAD-X actions with a cpymad.madx.Madx object, that are very specific to
-LHC and HLLHC use cases.
+what I do in LHC and HLLHC use cases.
 """
-from typing import Dict, List, Sequence, Tuple
+from typing import List, Sequence
 
 import numpy as np
-import tfs
 
 from cpymad.madx import Madx
 from loguru import logger
 
-from pyhdtoolkit.cpymadtools.constants import DEFAULT_TWISS_COLUMNS
-from pyhdtoolkit.cpymadtools.twiss import get_pattern_twiss
-
 # ----- Setup Utlites ----- #
 
 
-def make_lhc_beams(madx: Madx, energy: float = 6500, emittance: float = 3.75e-6, **kwargs) -> None:
+def make_lhc_beams(madx: Madx, energy: float = 7000, emittance: float = 3.75e-6, **kwargs) -> None:
     """
     Define beams with default configuratons for `LHCB1` and `LHCB2` sequences.
 
@@ -47,9 +43,10 @@ def make_lhc_beams(madx: Madx, energy: float = 6500, emittance: float = 3.75e-6,
             particle="proton",
             bv=1 if beam == 1 else -1,
             energy=energy,
-            npart=1.0e10,
+            npart=1.15e11,
             ex=geometric_emit,
             ey=geometric_emit,
+            sige=4.5e-4,
             **kwargs,
         )
 
@@ -137,7 +134,14 @@ def apply_lhc_rigidity_waist_shift_knob(
 ) -> None:
     """
     Applies the LHC rigidity waist shift knob, moving the waist left or right of IP. If you don't know what
-    this is, you should not be using this function.
+    this is, you should not be using this function. The waist shift is done by unbalancing the
+    triplet powering knob of the left and right-hand sides of the IP.
+
+    Warning: Applying the shift will modify your tunes and most likely flip them, making a subsequent
+    matching impossible if your lattice has coupling. To avoid this, match to tunes split further apart
+    before applying the waist shift knob, and then match to the desired working point. For instance for
+    the LHC, matching to (62.27, 60.36) before applying and afterwards rematching to (62.31, 60.32) usually
+    works well.
 
     Args:
         madx (cpymad.madx.Madx): an instanciated cpymad Madx object.
@@ -150,16 +154,15 @@ def apply_lhc_rigidity_waist_shift_knob(
     """
     logger.info(f"Applying Rigidity Waist Shift knob with a unit setting of {rigidty_waist_shift_value}")
     logger.warning("You should re-match tunes & chromaticities after this")
-    knob_variables = (f"kqx.r{ir:d}", f"kqx.l{ir:d}")  # Closest IP triplet
-    right_knob, left_knob = knob_variables
+    right_knob, left_knob = f"kqx.r{ir:d}", f"kqx.l{ir:d}"  # IP triplet default knob (no trims)
 
     current_right_knob = madx.globals[right_knob]
     current_left_knob = madx.globals[left_knob]
 
-    if side == "left":
+    if side.lower() == "left":
         madx.globals[right_knob] = (1 - rigidty_waist_shift_value * 0.005) * current_right_knob
         madx.globals[left_knob] = (1 + rigidty_waist_shift_value * 0.005) * current_left_knob
-    elif side == "right":
+    elif side.lower() == "right":
         madx.globals[right_knob] = (1 + rigidty_waist_shift_value * 0.005) * current_right_knob
         madx.globals[left_knob] = (1 - rigidty_waist_shift_value * 0.005) * current_left_knob
     else:
@@ -194,6 +197,204 @@ def apply_lhc_coupling_knob(
     logger.trace(f"Set '{knob_name}' to {madx.globals[knob_name]}")
 
 
+def install_ac_dipole_as_kicker(
+    madx: Madx,
+    deltaqx: float,
+    deltaqy: float,
+    sigma_x: float,
+    sigma_y: float,
+    beam: int = 1,
+    start_turn: int = 100,
+    ramp_turns: int = 2000,
+    top_turns: int = 6600,
+) -> None:
+    """
+    Installs an AC dipole for (HL)LHC BEAM 1 OR 2 ONLY.
+    The AC Dipole does impact the orbit as well as the betatron functions when turned on. Unfortunately in
+    MAD-X, it cannot be modeled to do both at the same time. This routine introduces an AC Dipole as a
+    kicker element so that its effect can be seen on particle orbit in tracking. It DOES NOT affect TWISS
+    functions. See https://journals.aps.org/prab/abstract/10.1103/PhysRevSTAB.11.084002 (part III).
+
+    This function assumes that you have already defined lhcb1/lhcb2, made a beam for it (BEAM command or
+    `make_lhc_beams` function), matched to your desired working point and made a TWISS.
+
+    Args:
+        madx (cpymad.madx.Madx): an instanciated cpymad Madx object.
+        deltaqx (float): the deltaQx (horizontal tune excitation) used by the AC dipole.
+        deltaqy (float): the deltaQy (vertical tune excitation) used by the AC dipole.
+        sigma_x (float): the horizontal amplitude to drive the beam to, in bunch sigma.
+        sigma_y (float): the vertical amplitude to drive the beam to, in bunch sigma.
+        beam (int): the LHC beam to install the AC Dipole into, either 1 or 2. Defaults to 1.
+        start_turn (int): the turn at which to start ramping up the AC dipole. Defaults to 100.
+        ramp_turns (int): the number of turns to use for the ramp-up and the ramp-down of the AC dipole.
+            This number is important in order to preserve the adiabaticity of the cycle. Defaults to 2000
+            as in the LHC.
+        top_turns (int): the number of turns to drive the beam for. Defaults to 6600 as in the LHC.
+    """
+    logger.warning("This AC Dipole is implemented as a kicker and will not affect TWISS functions!")
+    logger.info("This routine should be done after 'match', 'twiss' and 'makethin' for the appropriate beam")
+
+    if top_turns > 6600:
+        logger.warning(
+            f"Configuring the AC Dipole for {top_turns:d} of driving is fine for MAD-X but is "
+            "higher than what the device can do in the (HL)LHC! Beware."
+        )
+    ramp1, ramp2 = start_turn, start_turn + ramp_turns
+    ramp3 = ramp2 + top_turns
+    ramp4 = ramp3 + ramp_turns
+
+    logger.debug("Retrieving tunes from internal tables")
+    q1, q2 = madx.table.summ.q1[0], madx.table.summ.q2[0]
+    logger.trace(f"Retrieved values are q1 = {q1:.5f}, q2 = {q2:.5f}")
+    q1_dipole, q2_dipole = q1 + deltaqx, q2 + deltaqy
+
+    logger.trace("Querying BETX and BETY at AC Dipole location")
+    # All below is done as model_creator macros with `.input()` calls
+    madx.input(f"pbeam = beam%lhcb{beam:d}->pc;")
+    madx.input(f"betxac = table(twiss, MKQA.6L4.B{beam:d}, BEAM, betx);")
+    madx.input(f"betyac = table(twiss, MKQA.6L4.B{beam:d}, BEAM, bety);")
+
+    logger.trace("Calculating AC Dipole voltage values")
+    madx.input(f"voltx = 0.042 * pbeam * ABS({deltaqx}) / SQRT(180.0 * betxac) * {sigma_x}")
+    madx.input(f"volty = 0.042 * pbeam * ABS({deltaqy}) / SQRT(177.0 * betyac) * {sigma_y}")
+
+    logger.trace("Defining kicker elements for transverse planes")
+    madx.input(
+        f"MKACH.6L4.B{beam:d}: hacdipole, l=0, freq:={q1_dipole}, lag=0, volt=voltx, ramp1={ramp1}, "
+        f"ramp2={ramp2}, ramp3={ramp3}, ramp4={ramp4};"
+    )
+    madx.input(
+        f"MKACV.6L4.B{beam:d}: vacdipole, l=0, freq:={q2_dipole}, lag=0, volt=volty, ramp1={ramp1}, "
+        f"ramp2={ramp2}, ramp3={ramp3}, ramp4={ramp4};"
+    )
+
+    logger.info(
+        f"Installing AC Dipole kicker with driven tunes of Qx_D = {q1_dipole:.5f}  |  Qy_D = {q2_dipole:.5f}"
+    )
+    madx.command.seqedit(sequence=f"lhcb{beam:d}")
+    madx.command.flatten()
+    # The kicker version is meant for a thin lattice and is installed a right at MKQA.6L4.B[12] (at=0)
+    madx.command.install(element=f"MKACH.6L4.B{beam:d}", at=0, from_=f"MKQA.6L4.B{beam:d}")
+    madx.command.install(element=f"MKACV.6L4.B{beam:d}", at=0, from_=f"MKQA.6L4.B{beam:d}")
+    madx.command.endedit()
+
+    logger.warning(
+        f"Sequence LHCB{beam:d} is now re-used for changes to take effect. Beware that this will reset it, "
+        "remove errors etc."
+    )
+    madx.use(sequence=f"lhcb{beam:d}")
+
+
+def install_ac_dipole_as_matrix(madx: Madx, deltaqx: float, deltaqy: float, beam: int = 1) -> None:
+    """
+    Installs an AC dipole as a matrix element for (HL)LHC BEAM 1 OR 2 ONLY.
+    The AC Dipole does impact the orbit as well as the betatron functions when turned on. Unfortunately in
+    MAD-X, it cannot be modeled to do both at the same time. This routine introduces an AC Dipole as a
+    matrix element so that its effect can be seen on the TWISS functions. It DOES NOT affect tracking.
+    See https://journals.aps.org/prab/abstract/10.1103/PhysRevSTAB.11.084002 (part III).
+
+    This function assumes that you have already defined lhcb1/lhcb2, made a beam for it (BEAM command or
+    `make_lhc_beams` function), matched to your desired working point and made a TWISS.
+
+    Args:
+        madx (cpymad.madx.Madx): an instanciated cpymad Madx object.
+        deltaqx (float): the deltaQx (horizontal tune excitation) used by the AC dipole.
+        deltaqy (float): the deltaQy (vertical tune excitation) used by the AC dipole.
+        beam (int): the LHC beam to install the AC Dipole into, either 1 or 2. Defaults to 1.
+    """
+    logger.warning("This AC Dipole is implemented as a matrix and will not affect particle tracking!")
+    logger.info("This routine should be done after 'match', 'twiss' and 'makethin' for the appropriate beam.")
+
+    logger.debug("Retrieving tunes from internal tables")
+    q1, q2 = madx.table.summ.q1[0], madx.table.summ.q2[0]
+    logger.trace(f"Retrieved values are q1 = {q1:.5f}, q2 = {q2:.5f}")
+    q1_dipole, q2_dipole = q1 + deltaqx, q2 + deltaqy
+
+    logger.trace("Querying BETX and BETY at AC Dipole location")
+    # All below is done as model_creator macros with `.input()` calls
+    madx.input(f"betxac = table(twiss, MKQA.6L4.B{beam:d}, BEAM, betx);")
+    madx.input(f"betyac = table(twiss, MKQA.6L4.B{beam:d}, BEAM, bety);")
+
+    logger.trace("Calculating AC Dipole matrix terms")
+    madx.input(f"hacmap21 = 2 * (cos(2*pi*{q1_dipole}) - cos(2*pi*{q1})) / (betxac * sin(2*pi*{q1}));")
+    madx.input(f"vacmap43 = 2 * (cos(2*pi*{q2_dipole}) - cos(2*pi*{q2})) / (betyac * sin(2*pi*{q2}));")
+
+    logger.trace("Defining matrix elements for transverse planes")
+    madx.input(f"hacmap: matrix, l=0, rm21=hacmap21;")
+    madx.input(f"vacmap: matrix, l=0, rm43=vacmap43;")
+
+    logger.info(
+        f"Installing AC Dipole matrix with driven tunes of Qx_D = {q1_dipole:.5f}  |  Qy_D = {q2_dipole:.5f}"
+    )
+    madx.command.seqedit(sequence=f"lhcb{beam:d}")
+    madx.command.flatten()
+    # The matrix version is meant for a thick lattice and is installed a little after MKQA.6L4.B[12]
+    madx.command.install(element="hacmap", at="1.583 / 2", from_=f"MKQA.6L4.B{beam:d}")
+    madx.command.install(element="vacmap", at="1.583 / 2", from_=f"MKQA.6L4.B{beam:d}")
+    madx.command.endedit()
+
+    logger.warning(
+        f"Sequence LHCB{beam:d} is now re-used for changes to take effect. Beware that this will reset it, "
+        "remove errors etc."
+    )
+    madx.use(sequence=f"lhcb{beam:d}")
+
+
+def vary_independent_ir_quadrupoles(
+    madx: Madx, quad_numbers: Sequence[int], ip: int, sides: Sequence[str] = ("r", "l"), beam: int = 1
+) -> None:
+    """
+    Send the `vary` commands for the desired quadrupoles in the IRs. The independent quadrupoles for which
+    this is implemented are Q4 to Q13 included. This is useful to setup some specific matching involving
+    these elements.
+
+    It is necessary to have defined a 'brho' variable when creating your beams.
+
+    Args:
+        madx (cpymad.madx.Madx): an instanciated cpymad Madx object.
+        quad_numbers (Sequence[int]): quadrupoles to be varied, by number (aka position from IP).
+        ip (int): the IP around which to apply the instructions. Defaults to 1.
+        sides (Sequence[str]): the sides of IP to act on. Should be `R` for right and `L` for left.
+            Defaults to both sides of the IP.
+        beam (int): the beam for which to apply the instructions. Defaults to 1.
+    """
+    if (
+        ip not in (1, 2, 5, 8)
+        or any(side.upper() not in ("R", "L") for side in sides)
+        or any(quad not in (4, 5, 6, 7, 8, 9, 10, 11, 12, 13) for quad in quad_numbers)
+    ):
+        logger.error("Either the IP number of the side provided are invalid, not applying any error.")
+        raise ValueError("Invalid 'quad_numbers', 'ip', 'sides' argument")
+
+    logger.debug(f"Preparing a knob involving quadrupoles {quad_numbers}")
+    # Each quad has a specific power circuit used for their k1 boundaries
+    power_circuits: Dict[int, str] = {
+        4: "mqy",
+        5: "mqml",
+        6: "mqml",
+        7: "mqm",
+        8: "mqml",
+        9: "mqm",
+        10: "mqml",
+        11: "mqtli",
+        12: "mqt",
+        13: "mqt",
+    }
+    for quad in quad_numbers:
+        circuit = power_circuits[quad]
+        for side in sides:
+            logger.trace(f"Sending vary command for Q{quad}{side.upper()}{ip}")
+            madx.command.vary(
+                name=f"kq{'t' if quad >= 11 else ''}{'l' if quad == 11 else ''}{quad}.{side}{ip}b{beam}",
+                step=1e-7,
+                lower=f"-{circuit}.{'b' if quad == 7 else ''}{quad}{side}{ip}.b{beam}->kmax/brho",
+                upper=f"+{circuit}.{'b' if quad == 7 else ''}{quad}{side}{ip}.b{beam}->kmax/brho",
+            )
+
+
+# ----- Output Utilities ----- #
+
+
 def make_sixtrack_output(madx: Madx, energy: int) -> None:
     """
     INITIAL IMPLEMENTATION CREDITS GO TO JOSCHUA DILLY (@JoschD).
@@ -215,86 +416,6 @@ def make_sixtrack_output(madx: Madx, energy: int) -> None:
     madx.sixtrack(cavall=True, radius=0.017)  # this value is only ok for HL(LHC) magnet radius
 
 
-def install_ac_dipole(
-    madx: Madx,
-    deltaqx: float,
-    deltaqy: float,
-    sigma_x: float,
-    sigma_y: float,
-    geometric_emit: float = None,
-    start_turn: int = 100,
-    ramp_turns: int = 2000,
-    top_turns: int = 6600,
-) -> None:
-    """
-    Installs an AC dipole for BEAM 1 ONLY.
-    This function assumes that you have already defined lhcb1, made a beam for it (BEAM command or
-    `make_lhc_beams` function) and matched to your desired working point.
-
-    Args:
-        madx (cpymad.madx.Madx): an instanciated cpymad Madx object.
-        deltaqx (float): the deltaQx (horizontal tune excitation) used by the AC dipole.
-        deltaqy (float): the deltaQy (vertical tune excitation) used by the AC dipole.
-        sigma_x (float): the horizontal amplitude to drive the beam to, in bunch sigma.
-        sigma_y (float): the vertical amplitude to drive the beam to, in bunch sigma.
-        geometric_emit (float): the geometric emittance that was used when defining the beam. If not
-            provided, it is assumed that 'geometric_emit' is a defined global in MAD-X, and the value will
-            be directly queried from the internal tables.
-        start_turn (int): the turn at which to start ramping up the AC dipole. Defaults to 100.
-        ramp_turns (int): the number of turns to use for the ramp-up and the ramp-down of the AC dipole.
-            This number is important in order to preserve the adiabaticity of the cycle. Defaults to 2000
-            as in the LHC.
-        top_turns (int): the number of turns to drive the beam for. Defaults to 6600 as in the LHC.
-    """
-    if top_turns > 6600:
-        logger.warning(
-            f"Configuring the AC Dipole for {top_turns} of driving is fine for MAD-X but is "
-            "higher than what the device can do in the (HL)LHC! Beware."
-        )
-    ramp1, ramp2 = start_turn, start_turn + ramp_turns
-    ramp3 = ramp2 + top_turns
-    ramp4 = ramp3 + ramp_turns
-
-    logger.debug("Retrieving tunes from internal tables")
-    q1, q2 = madx.table.summ.q1[0], madx.table.summ.q2[0]
-    logger.trace(f"Retrieved values are q1 = {q1}, q2 = {q2}")
-    q1_dipole, q2_dipole = q1 + deltaqx, q2 + deltaqy
-
-    if not geometric_emit:
-        logger.debug(
-            f"No value provided for the geometric emittance used when creating the beam, the value will be "
-            f"queried from MAD-X's global 'geometric_emit'"
-        )
-        geometric_emit = madx.globals["geometric_emit"]
-
-    logger.info(f"Installing AC Dipole to drive the tunes to Qx_D = {q1_dipole}  |  Qy_D = {q2_dipole}")
-    madx.input(
-        f"MKACH.6L4.B1: hacdipole, l=0, freq:={q1_dipole}, lag=0, volt:=voltx, ramp1={ramp1}, "
-        f"ramp2={ramp2}, ramp3={ramp3}, ramp4={ramp4};"
-    )
-    madx.input(
-        f"MKACV.6L4.B1: vacdipole, l=0, freq:={q2_dipole}, lag=0, volt:=volty, ramp1={ramp1}, "
-        f"ramp2={ramp2}, ramp3={ramp3}, ramp4={ramp4};"
-    )
-    madx.command.seqedit(sequence="lhcb1")
-    madx.command.flatten()
-    madx.command.install(element="MKACH.6L4.B1", at="0.0", from_="MKQA.6L4.B1")
-    madx.command.install(element="MKACV.6L4.B1", at="0.0", from_="MKQA.6L4.B1")
-    madx.command.endedit()
-
-    logger.trace("Querying BETX and BETY at AC Dipole location")
-    madx.input("betax_acd = table(twiss, MKQA.6L4.B1, betx);")
-    madx.input("betay_acd = table(twiss, MKQA.6L4.B1, bety);")
-
-    betax_acd = madx.globals["betax_acd"]
-    betay_acd = madx.globals["betay_acd"]
-    brho = madx.sequence.lhcb1.beam.brho
-    voltx = sigma_x * np.sqrt(geometric_emit) * brho * np.abs(deltaqx) * 4 * np.pi / np.sqrt(betax_acd)
-    volty = sigma_y * np.sqrt(geometric_emit) * brho * np.abs(deltaqy) * 4 * np.pi / np.sqrt(betay_acd)
-    madx.globals["voltx"] = voltx
-    madx.globals["volty"] = volty
-
-
 # ----- Miscellaneous Utilities ----- #
 
 
@@ -345,64 +466,51 @@ def make_lhc_thin(madx: Madx, sequence: str, slicefactor: int = 1, **kwargs) ->
 
     madx.use(sequence=sequence)
     style = kwargs.get("style", "teapot")
-    makedipedge = kwargs.get("makedipedge", True)
+    makedipedge = kwargs.get("makedipedge", False)  # defaults to False to compensate default TEAPOT style
     madx.command.makethin(sequence=sequence, style=style, makedipedge=makedipedge)
 
 
-# ----- Twiss Utilities ----- #
-
-
-def get_ips_twiss(madx: Madx, columns: Sequence[str] = DEFAULT_TWISS_COLUMNS, **kwargs) -> tfs.TfsDataFrame:
+def re_cycle_sequence(madx: Madx, sequence: str = "lhcb1", start: str = "IP3") -> None:
     """
-    Quickly get the `TWISS` table for certain variables at IP locations only. The `SUMM` table will be
-    included as the TfsDataFrame's header dictionary.
+    Re-cycle the provided sequence from a different starting point.
+
+    Args:
+        madx (Madx): an instantiated cpymad.madx.Madx object.
+        sequence (str): the sequence to re cycle.
+        start (str): element to start the new cycle from.
+    """
+    logger.debug(f"Re-cycling sequence '{sequence}' from {start}")
+    madx.command.seqedit(sequence=sequence)
+    madx.command.flatten()
+    madx.command.cycle(start=start)
+    madx.command.endedit()
+
+
+def match_no_coupling_through_ripkens(
+    madx: Madx, sequence: str = None, location: str = None, vary_knobs: Sequence[str] = None
+) -> None:
+    """
+    Matching routine to get cross-term Ripken parameters beta_12 and beta_21 to be 0 at a given location.
 
     Args:
         madx (cpymad.madx.Madx): an instanciated cpymad Madx object.
-        columns (Sequence[str]): the variables to be returned, as columns in the DataFrame.
-
-    Keyword Args:
-        Any keyword argument that can be given to the MAD-X TWISS command, such as `chrom`, `ripken`,
-        `centre` or starting coordinates with `betax`, 'betay` etc.
-
-    Returns:
-        A TfsDataFrame of the twiss output.
+        sequence (str): name of the sequence to activate for the matching.
+        location (str): the name of the element at which one wants the cross-term Ripkens to be 0.
+        vary_knobs (Sequence[str]): the variables names to 'vary' in the MADX routine.
     """
-    logger.info("Getting Twiss at IPs")
-    return get_pattern_twiss(madx=madx, patterns=["IP"], columns=columns, **kwargs)
+    logger.info(f"Matching Ripken parameters for no coupling at location {location}")
+    logger.debug("Creating macro tu update Ripkens")
+    madx.input("do_ripken: macro = {twiss, ripken=True;}")  # cpymad needs .input for macros
 
-
-def get_ir_twiss(
-    madx: Madx, ir: int, columns: Sequence[str] = DEFAULT_TWISS_COLUMNS, **kwargs
-) -> tfs.TfsDataFrame:
-    """
-    Quickly get the `TWISS` table for certain variables for one IR, meaning at the IP and Q1 to Q3 both
-    left and right of the IP. The `SUMM` table will be included as the TfsDataFrame's header dictionary.
-
-    Args:
-        madx (cpymad.madx.Madx): an instanciated cpymad Madx object.
-        ir (int): which interaction region to get the TWISS for.
-        columns (Sequence[str]): the variables to be returned, as columns in the DataFrame.
-
-    Keyword Args:
-        Any keyword argument that can be given to the MAD-X TWISS command, such as `chrom`, `ripken`,
-        `centre` or starting coordinates with `betax`, 'betay` etc.
-
-    Returns:
-        A TfsDataFrame of the twiss output.
-    """
-    logger.info(f"Getting Twiss for IR {ir:d}")
-    return get_pattern_twiss(
-        madx=madx,
-        patterns=[
-            f"IP{ir:d}",
-            f"MQXA.[12345][RL]{ir:d}",
-            f"MQXB.[AB][12345][RL]{ir:d}",
-            f"MQXF[AB].[AB][12345][RL]{ir:d}",  # this is for HLLHC
-        ],
-        columns=columns,
-        **kwargs,
-    )
+    logger.debug("Matching Parameters")
+    madx.command.match(sequence=sequence, use_macro=True, chrom=True)
+    for knob in vary_knobs:
+        madx.command.vary(name=knob)
+    madx.command.use_macro(name="do_ripken")
+    madx.input(f"constraint, expr=table(twiss, {location}, beta12)=0")  # need input else includes " and fails
+    madx.input(f"constraint, expr=table(twiss, {location}, beta21)=0")  # need input else includes " and fails
+    madx.command.lmdif(calls=500, tolerance=1e-21)
+    madx.command.endmatch()
 
 
 # ----- Helpers ----- #
