@@ -7,4 +7,6 @@ These are miscellaneous utilities to perform optics calculation from simulation
 :license: MIT, see LICENSE for more details.
 """
 
-from .beam import Beam
+from .beam import Beam, compute_beam_parameters
+from .ripken import lebedev_beam_size
+from .twiss import courant_snyder_transform
